/*
 * Raima Database Manager
 *
 * Copyright (c) 2019 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 */

/*! \file
 * \brief Source code for the RDM Perf08 example
 */
#include <stdio.h>
#include <string.h>
#include "perftimer.h"
#include "print_error.h"
#include "rdm.h"

/* Generated \c struct and \c typedef definitions to be used with the RDM APIs */
#include "perf08_db_structs.h"

/* Generated catalog definition to be used with the RDM rdm_dbSetCatalog() API */
#include "perf08_db_cat.h"

/*! Number of levels in the BOM */
#define NUM_LEVELS 6

/*! Number of members per level in the BOM */
#define NUM_MEMBERS 8

/*! Context structure for the Perf08 example */
typedef struct _PERF08_CTX
{
    int max_level;   /*!< Maximum BOM level */
    int max_members; /*!< Maximum number of members per BOM level */
    int count;       /*!< Count of BOM members */
    RDM_TFS hTFS;    /*!< RDM TFS handle */
    RDM_DB hDB;      /*!< RDM DB handle */
} PERF08_CTX;

/*!
 * \mainpage Perf08 Popcorn Example
 *
 * The Perf08 example creates and reads a Bill-of-Materials using relational
 * index definitions.
 *
 * Each time you run this example
 * \li The database is initialized (removing all existing data)
 * \li A new Bill-of-Materials is created
 * \li The Bill-of-Materials is read
 * \li Database is cleaned up
 *
 * \par Table of Contents
 *
 * - \subpage hDB
 * - \subpage hRDMfunc
 * - \subpage hPGMfunc
 *
 * For additional information, please refer to the product documentation at http://docs.raima.com/.
 *
 * \page hDB Database Schema
 *
 * \par Database Schema Definition
 *
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file perf08_db.sdl. The example utilizes
 * the RDM Core DDL syntax and specifies a database suitable for storing a
 * Bill-of-Materials.
 *
 * \include perf08_db.sdl
 *
 * The schema was compiled using the RDM ddlp utility with the -c option to
 * generate an embedded database dictionary (used by the RDM rdm_dbSetCatalog API).
 *
 * \code rdm-compile -a -s perf08_db.sdl \endcode
 *
 * The database dictionary information is embedded inside the application so
 * an external catalog file stored on the file system is not required.
 *
 * \page hRDMfunc RDM Functions Used
 * The following RDM functions are illustrated:
 * \li rdm_rdmAllocTFS/rdm_tfsFree
 * \li rdm_tfsAllocDatabase/rdm_dbFree
 * \li rdm_dbSetCatalog
 * \li rdm_dbOpen/rdm_dbClose
 * \li rdm_dbInsertRow
 * \li rdm_cursorFree
 * \li rdm_dbGetRowsByKeyInRangeKeyRange
 * \li rdm_dbGetRows
 * \li rdm_cursorMoveToFirst/rdm_cursorMoveToNext
 * \li rdm_cursorReadColumn
 * \li rdm_dbDeleteAllRowsFromDatabase
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 *
 * \page hPGMfunc Program Functions
 * \li open_and_initialize() - \copybrief open_and_initialize
 * \li bom_random_id() - \copybrief bom_random_id
 * \li build_bill() - \copybrief build_bill
 * \li create_bom() - \copybrief create_bom
 * \li get_cost() - \copybrief get_cost
 * \li rollup_bom() - \copybrief rollup_bom
 * \li cleanup() - \copybrief cleanup
 * \li main() - \copybrief main
 */

/*! \brief Initialize the RDM runtime library for use in the perf08 example
 *
 * This function initializes the RDM Transactional File Server (TFS) to use
 * the EMBEDED TFS implementation.  It also allocates a database handle
 * and opens the "perf08" database in exclusive mode.
 * Exclusve mode does not require database locks or transactions.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
RDM_RETCODE open_and_initialize (PERF08_CTX *ctx) /**< [in/out] Pointer to the application context */
{
    RDM_RETCODE rc;
    perfTimer_t timer;

    timeMeasureBegin (&timer);

    rc = rdm_rdmAllocTFS ("tfstype=EMBED;docroot=.", &ctx->hTFS);
    print_error (rc);

    if (rc == sOKAY)
    {
        /* If we have a TFS handle ... */
        rc = rdm_tfsAllocDatabase (ctx->hTFS, &ctx->hDB);
        print_error (rc);

        if (rc == sOKAY)
        {
            /* If we have a DB handle ... */
            rc = rdm_dbSetCatalog (ctx->hDB, perf08_db_cat);
            print_error (rc);

            if (rc == sOKAY)
            {
                rc = rdm_dbSetOptions (ctx->hDB, "storage=ondisk;durability=consistent");
                print_error (rc);
            }
            if (rc == sOKAY)
            {
                rc = rdm_dbOpen (ctx->hDB, "perf08", RDM_OPEN_EXCLUSIVE);
                print_error (rc);
            }
            if (rc == sOKAY)
            {
                rc = rdm_dbDeleteAllRowsFromDatabase (ctx->hDB);
                print_error (rc);
            }

            /* If any of the above DB operations failed, free the DB handle */
            if (rc != sOKAY)
            {
                rdm_dbFree (ctx->hDB);
            }
        }

        /* If any of the above DB operations failed, free the TFS handle */
        if (rc != sOKAY)
        {
            rdm_tfsFree (ctx->hTFS);
        }
    }

    timeMeasureEnd (&timer);

    if (rc == sOKAY)
    {
        /* returns with hTFS and hDB allocated if sOKAY */
        /* or nothing allocated if not sOKAY */

        ctx->max_level = NUM_LEVELS;
        ctx->max_members = NUM_MEMBERS;

        /* Display information */
        printf ("\n");
        printf ("Preparing a new database:\t\t\t%u milliseconds\n", timeMeasureDiff (&timer));
        printf ("\tDisk-based\n");
        printf ("\tOpened exclusively\n");
    }
    return rc;
}

/*! \brief Cleanup the RDM runtime library
 *
 * This functions closes all open databases and cleans up the RDM db handle
 * used in the perf08 example.  It also terminates the TFS used
 * in the example.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
void cleanup (PERF08_CTX *ctx) /**< [in] Pointer to the application context */
{
    perfTimer_t timer;

    timeMeasureBegin (&timer);

    /* Close any open databases */
    rdm_dbClose (ctx->hDB);

    /* Close the task */
    rdm_dbFree (ctx->hDB);

    /* Terminate the TFS */
    rdm_tfsFree (ctx->hTFS);

    timeMeasureEnd (&timer);

    printf ("\n");
    printf ("Cleanup:\t\t\t\t\t%u milliseconds\n", timeMeasureDiff (&timer));
}

/*! \brief Generate a 15-character alpha-numeric part id
 *
 * This function generates a part id for an item
 *
 */
void bom_random_id (PERF08_CTX *ctx, /**< [in/out] Pointer to the application context */
                    char *string)    /**< [out] generated id_code */
{
    sprintf (string, "%-15.15d", ++ctx->count);
}

/*! \brief Recursive rountine to build one level of a bill by adding components to a parent
 *
 * This function recursively builds a full bill of materials in the perf08
 * database.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
RDM_RETCODE build_bill (PERF08_CTX *ctx,            /**< [in/out] Pointer to the application context */
                        RDM_CURSOR cursorOwnerItem, /**< [in] Owner cursor of the current level */
                        const char *owner_id_code,  /**< [in] Owner ID code */
                        int current_level)          /**< [in] Current level */
{
    RDM_RETCODE rc = sOKAY;

    if (current_level < ctx->max_level)
    {
        RDM_CURSOR cursorItem = NULL;
        RDM_CURSOR cursorBill = NULL;
        int ii;

        for (ii = 0; rc == sOKAY && ii < ctx->max_members; ii++)
        {
            /* create new TABLE_ITEM record, for component item */
            ITEM item_rec;

            item_rec.COST = 1.0L;
            item_rec.DESCRIPTION[0] = '\0';
            bom_random_id (ctx, item_rec.ID_CODE);
            rc = rdm_dbInsertRow (ctx->hDB, TABLE_ITEM, &item_rec, sizeof (item_rec), &cursorItem);

            /* create new TABLE_BILL record */

            if (rc == sOKAY)
            {
                BILL bill_rec;

                bill_rec.QUANTITY = 1.0L;
                bill_rec.LEVEL = current_level;
                strncpy (bill_rec.BOM_ID_CODE, owner_id_code, sizeof (bill_rec.BOM_ID_CODE));
                strncpy (bill_rec.WHERE_USED_ID_CODE, item_rec.ID_CODE, sizeof (bill_rec.WHERE_USED_ID_CODE));
                rc = rdm_dbInsertRow (ctx->hDB, TABLE_BILL, &bill_rec, sizeof (bill_rec), &cursorBill);
            }

            /* Do a recursiv call to build next level */
            if (rc == sOKAY)
            {
                rc = build_bill (ctx, cursorItem, item_rec.ID_CODE, current_level + 1);
            }
        }

        if (cursorItem)
        {
            rdm_cursorFree (cursorItem);
        }

        if (cursorBill)
        {
            rdm_cursorFree (cursorBill);
        }
    }

    return rc;
}

/*! \brief Top level function to create a BOM
 *
 * This function recursively builds a full bill of materials in the perf08
 * database.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
RDM_RETCODE create_bom (PERF08_CTX *ctx) /**< [in/out] Pointer to the application context */
{
    RDM_RETCODE rc;
    perfTimer_t timer;

    timeMeasureBegin (&timer);

    RDM_CURSOR cursorRootItem = NULL;
    ITEM item_rec;

    item_rec.COST = 1.0L;
    item_rec.DESCRIPTION[0] = '\0';
    bom_random_id (ctx, item_rec.ID_CODE);

    /* create top level TABLE_ITEM record */
    rc = rdm_dbInsertRow (ctx->hDB, TABLE_ITEM, &item_rec, sizeof (item_rec), &cursorRootItem);

    if (rc == sOKAY)
    {
        /* build multi-level bill */
        rc = build_bill (ctx, cursorRootItem, item_rec.ID_CODE, 0);
    }

    timeMeasureEnd (&timer);
    printf ("\t\t%u milliseconds\n", timeMeasureDiff (&timer));
    return rc;
}

/*! \brief Recursive rountine roll up cost from lower levels of a BOM
 *
 * This function recursively rolls-up the cost of a BOM.  The costs are
 * only stored at the lowest levels of the BOM.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
RDM_RETCODE get_cost (PERF08_CTX *ctx,            /**< [in]  Pointer to the application context */
                      RDM_CURSOR cursorOwnerItem, /**< [in] Owner cursor of the current level */
                      double *total_cost)         /**< [out] Total cost accummulated */
{
    RDM_RETCODE rc;
    RDM_CURSOR cursorBill = NULL;
    double dTotalCost = 0.0L;
    char id_code[RDM_COLUMN_SIZE (ITEM, ID_CODE)];
    RDM_RANGE_KEY id_code_key = {id_code, 0, 1, RDM_RANGE_OPEN};

    /* get number of components connected to this TABLE_ITEM record */
    rc = rdm_cursorReadColumn (cursorOwnerItem, COL_ITEM_ID_CODE, &id_code, sizeof (id_code), NULL);

    if (rc == sOKAY)
    {
        rc =
            rdm_dbGetRowsByKeyInRangeKeyRange (ctx->hDB, KEY_BILL_BOM_ID_CODE, &id_code_key, &id_code_key, &cursorBill);
    }

    if (rc == sOKAY)
    {
        rc = rdm_cursorMoveToFirst (cursorBill);

        if (rc == sENDOFCURSOR)
        {
            rc = rdm_cursorReadColumn (cursorOwnerItem, COL_ITEM_COST, &dTotalCost, sizeof (dTotalCost), NULL);
        }
        else
        {
            RDM_CURSOR cursorItem = NULL;

            while (rc == sOKAY)
            {
                double quantity;
                double dCost;

                rc = rdm_cursorReadColumn (cursorBill, COL_BILL_WHERE_USED_ID_CODE, &id_code, sizeof (id_code), NULL);

                if (rc == sOKAY)
                {
                    rc = rdm_dbGetRowsByKeyInRangeKeyRange (ctx->hDB, KEY_ITEM_ID_CODE, &id_code_key, &id_code_key,
                                                            &cursorItem);
                }

                rc = rdm_cursorMoveToFirst (cursorItem);

                if (rc == sOKAY)
                {
                    rc = get_cost (ctx, cursorItem, &dCost);
                }

                if (rc == sOKAY)
                {
                    rc = rdm_cursorReadColumn (cursorBill, COL_BILL_QUANTITY, &quantity, sizeof (quantity), NULL);
                }

                if (rc == sOKAY)
                {
                    dTotalCost += dCost * quantity;
                }

                if (rc == sOKAY)
                {
                    rc = rdm_cursorMoveToNext (cursorBill);
                }
            }
            if (rc == sENDOFCURSOR)
            {
                rc = sOKAY;
            }
            if (cursorItem)
            {
                rdm_cursorFree (cursorItem);
            }
        }
        rdm_cursorFree (cursorBill);
    }

    if (rc == sOKAY)
    {
        *total_cost = dTotalCost;
    }
    return rc;
}

/*! \brief Top level function for rolling the BOM
 *
 * This function recursively rolls up and calculate the full cost
 * of a BOM.
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
RDM_RETCODE rollup_bom (PERF08_CTX *ctx) /**< [in]  Pointer to the application context */
{
    RDM_RETCODE rc;
    RDM_CURSOR cursorRootItem = NULL;
    double total_cost = 0.0L;
    perfTimer_t timer;

    timeMeasureBegin (&timer);

    /* Navigate the top level TABLE_ITEM record */
    rc = rdm_dbGetRows (ctx->hDB, TABLE_ITEM, &cursorRootItem);

    if (rc == sOKAY)
    {
        rc = rdm_cursorMoveToFirst (cursorRootItem);

        if (rc == sOKAY)
        {
            rc = get_cost (ctx, cursorRootItem, &total_cost);
        }

        rdm_cursorFree (cursorRootItem);
    }
    timeMeasureEnd (&timer);

    printf ("\tTotal cost is $%10.2lf", total_cost);
    printf ("\t\t%u milliseconds\n", timeMeasureDiff (&timer));

    return rc;
}

/*! \brief Main function for perf08 example
 *
 * The function initializes the RDM environment and runs the insert operations.
 * over a number of transaction block sizes
 *
 * @return Returns RDM_RETCODE status code (sOKAY if successful)
 */
int main ()
{
    PERF08_CTX ctx = {0};
    RDM_RETCODE rc;

    /* Display the example title */
    printf ("%s", RDM_EXAMPLE_DESC ("Perf08"));

    /* Initialize the TFS, task, and open/initialize the database */
    rc = open_and_initialize (&ctx);
    if (rc == sOKAY)
    {
        /* Build the BOM */
        printf ("\nCreating BOM with %d levels and %d members per level\n", ctx.max_level, ctx.max_members);
        printf ("\tBuilding bill of materials");
        create_bom (&ctx);

        /* Read the BOM - rollup cost */
        printf ("\nRolling up BOM cost\n");
        rollup_bom (&ctx);

        cleanup (&ctx);
    }
    return (int) rc;
}
