/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2010 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * ----------------------------------------------------------------------------
 */

/** \file rdmtimeapi.h
 *  \brief Header for the public Time API
 */

#include "rdmdatetimetypes.h"
#include "rdmretcodetypes.h"

#if defined(RDM_UNICODE)
#define rdm_timeFromString RDM_T_FUNC (rdm_timeFromString)
#define rdm_timeToString RDM_T_FUNC (rdm_timeToString)
#define rdm_timeNowAsString RDM_T_FUNC (rdm_timeNowAsString)
#endif

/** \defgroup time Time API
 * \ingroup datetime
 *
 * The RDM time datatype API.  The functions here are located in \lib{base}
 *
 * \addtogroup time
 * @{
 */

#if defined(__cplusplus)
extern "C" {
#endif

/** \brief Convert a string to local time
 *
 *  Convert a string to a packed local time.
 *
 *  \retval{sOKAY}
 *  \retval{eINVLITERAL}
 *  \retval{eRANGE}
 */
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_timeFromString (
    const RDM_TCHAR_T *str, /**< [in] The string to be converted */
    RDM_PACKED_TIME_T *ptm  /**< [out] The packed local time value. The buffer
                               need to be allocated by the caller. */
);

/** \brief Get the hour of the day
 *
 *  Get the hour from a packed local time.
 *
 *  \return The hour of the day
 */
RDM_EXPORT uint16_t EXTERNAL_FCN rdm_timeHour (
    RDM_PACKED_TIME_T tm /**< [in]  RDM packed local time value */
);

/** \brief Get the minute of the hour
 *
 *  Get the minute of the hour from a packed local time.
 *
 *  \return The minute of the hour
 */
RDM_EXPORT uint16_t EXTERNAL_FCN rdm_timeMinute (
    RDM_PACKED_TIME_T tm /**< [in]  RDM packed local time value */
);

/** \brief Get the second of the minute
 *
 *  Get the second of the minute from a packed local time.
 *
 *  \return The second of the minute
 */
RDM_EXPORT uint16_t EXTERNAL_FCN rdm_timeSecond (
    RDM_PACKED_TIME_T tm /**< [in]  RDM packed local time value */
);

/** \brief Get the fractions of the second
 *
 *  Get the fractions of the second in 1/10,000th of a second from a packed local time
 *
 *  The fractions of the second (0 - 9999)
 */
RDM_EXPORT uint16_t EXTERNAL_FCN rdm_timeFraction (
    RDM_PACKED_TIME_T tm /**< [in]  RDM packed local time value */
);

/** \brief Convert a time to a string
 *
 *  Convert a packed local time to a string.
 *
 *  \retval{sOKAY}
 *  \retval{eINVLITERAL}
 *  \retval{sTRUNCATE}
 */
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_timeToString (
    RDM_PACKED_TIME_T timeVal, /**< [in]  RDM time value */
    RDM_TIME_FORMAT time_fmt,  /**< [in]  RDM time format */
    RDM_TCHAR_T *buf,          /**< [out] result time string */
    size_t bufSize, /**< [in]  buffer size in number of ASCII characters it can hold */
    size_t *puSize /**< [out] resulting size in number of characters */
);

/** \brief Get the local time at the start of a day
 *
 *  Get the packed local time at the start of the day (midnight).
 *
 *  \return The local time at the start of the day
 */
RDM_EXPORT RDM_PACKED_TIME_T EXTERNAL_FCN rdm_timeZero (void);

/** \brief Get the current local time
 *
 *  Get the current local time for the timezone passed in.
 *
 *  \retval{sOKAY}
 *  \retval{eBASE_INVTIMEZONE}
 */
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_timeNow (
    int16_t time_zone,     /**< [in]  timezone value (in min) */
    RDM_PACKED_TIME_T *ptm /**< [out] current time adjusted with time_zone.  The
                              buffer need to be allocated by the caller. */
);

/** \brief Get current local time as a string
 *
 *  Get the current local time in string format.
 *
 *  \retval{sOKAY}
 *  \retval{eBASE_INVTIMEZONE}
 *  \retval{sTRUNCATE}
 */
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_timeNowAsString (
    int16_t time_zone,    /**< [in]  timezone value (in min) */
    RDM_TCHAR_T *timebuf, /**< [out] NOW value as string */
    size_t buflen /**< [in]  Size of timebuf in number of ASCII characters it can hold */
);

#if defined(__cplusplus)
}
#endif

/** @}*/
