/* ----------------------------------------------------------------------------
 * Raima Database Manager
 *
 * Copyright (c) 2014 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 * ----------------------------------------------------------------------------
 */

/** \file rdmtransapi.h
 *  \brief Header for the RDM Transaction APIs
 */

#include "rdmtypes.h"

/**
 *
 * \defgroup rdm_trans RDM Transaction APIs
 * \ingroup db_locking
 *
 *  See \ref locking for a more detailed description.
 *
 *  The core database transaction API.  These functions are located in \lib{rdm}
 *
 *  A transaction handle can be instantiated using ::rdm_dbStartUpdate,
 * ::rdm_dbStartRead, or ::rdm_dbStartSnapshot.
 *
 * \addtogroup rdm_trans
 * @{*/

#if defined(__cplusplus)
extern "C" {
#endif

/** \brief End a transactional operation

    rdm_transEnd() will process a transaction point. If the transaction
    point is the root then the transaction will be committed and the changes
    made a part of the database. If the transaction point is not the root then
    the transaction point and any subsequent transaction points will be deleted.
    If a handle to subsequent transaction points was kept it should be discarded
    as the used of an invalided transaction handle is undefined.

    Any read locks that were granted when or after the specified transaction
    point was created will be freed.

    See \ref locking for a more detailed description.

    \retval{sOKAY}

    \see rdm_dbStartUpdate
    \see rdm_dbStartRead
    \see rdm_dbStartSnapshot
    \see rdm_dbEnd
    \see rdm_transFree
*/
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_transEnd (
    RDM_TRANS trans /**< [in] A valid RDM transaction handle */
);

/** \brief End a transactional operation with a rollback

    rdm_dbEndRollback() will rollback an update transaction up to the specified
    transaction point. If the transaction point is the root then the
    transaction will be aborted and all locks will be freed. If the transaction
    point is not the root then the transaction point and any subsequent
    transaction points will be invalidated. If a handle to subsequent
    transaction points was kept it should be discarded as the used of an
    invalided transaction handle is undefined.

    See \ref locking for a more detailed description.

    \retval{sOKAY}
    \retval{eNOUPDATE}

    \see rdm_dbStartUpdate
    \see rdm_dbStartRead
    \see rdm_dbStartSnapshot
    \see rdm_transPrecommit
    \see rdm_transRollback
    \see rdm_dbEndRollback
    \see rdm_transFree
    */
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_transEndRollback (
    RDM_TRANS trans /**< [in] A valid RDM transaction handle */
);

/** \brief Free a transaction handle

    rdm_transFree() will free a \a hTrans without freeing any of the locks
    associated with the transaction point. If the specified
    transaction is a root transaction then the only way to end (commit/rollback)
    the transaction will be to use the rdm_dbEnd() or rdm_dbEndRollback()
    APIs. No transaction handle prior or subsequent to \a hTrans will be affected.

    See \ref locking for a more detailed description.

    \retval{sOKAY}

    \see rdm_dbStartUpdate
    \see rdm_dbStartRead
    \see rdm_dbStartSnapshot
    \see rdm_transEnd
    \see rdm_transEndRollback
*/
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_transFree (
    RDM_TRANS hTrans /**< [in] A valid RDM transaction handle */
);

/** \brief Precommit a transactional operation (Two-Phase Commit phase 1)

    rdm_transPrecommit() performs phase one of a two-phase commit. The function
   prepares to commit a transaction by writing to the database log file all
   changes made for the specified transaction. The transaction must be a root
   transaction for a database and not a transaction mark point. Soon after this
   call, rdm_dbEnd(), rdm_transEnd(), rdm_dbEndRollback() or
   rdm_transEndRollback() should be called to finally commit or rollback the
   transaction. The difference between calling rdm_transPrecommit() prior to
   rdm_dbEnd() or rdm_transEnd() is that all file I/O necessary to prepare the
   transaction will be done during the precommit, and all rdm_dbEnd() or
   rdm_transEnd() has left to do is a minimal task, essentially flipping a
   switch to commit the prepared transaction.

    \note This call does not change the status of locks, so a database should
   not be held in this state indefinitely.

    \retval{sOKAY}

    \see rdm_dbStartUpdate
    \see rdm_dbStartRead
    \see rdm_dbStartSnapshot
    \see rdm_dbPrecommit
    \see rdm_transEnd
    \see rdm_transEndRollback
*/
RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_transPrecommit (
    RDM_TRANS trans /**< [in] A valid RDM transaction handle */
);

/** \brief Rollback transactional operations

    rdm_transRollback() will rollback an update transaction up to the specified
    transaction point. All changes made to the database will be reverted,
    but any locks granted when the \a hTrans was started will remain. Locks
    that were granted to any transactions nested from \a hTrans will be freed.
    If a handle to subsequent transaction points was kept it should be discarded
    as the used of an invalided transaction handle is undefined.

    \a hTrans will remain and updates can continue to be applied to it.

    See \ref locking for a more detailed description.

    \retval{sOKAY}
    \retval{eNOUPDATE}

    \see rdm_dbStartUpdate
    \see rdm_dbStartRead
    \see rdm_dbStartSnapshot
    \see rdm_transEndRollback
*/

RDM_EXPORT RDM_RETCODE EXTERNAL_FCN rdm_transRollback (
    RDM_TRANS hTrans /**< [in] A valid RDM transaction handle */
);

#if defined(__cplusplus)
}
#endif

/** @}*/
